#!/usr/bin/env python
#
# Simple Marathon shapes-patcher utility
# by Loren Petrich
#
# Its patch format is the Marathon shapes-file format
#
# Works from Tito dal Canton's documentation:
# http://www.dalcanton.it/tito/hp/marathonshapes/

import sys, struct, os, hashlib, pipes

usage = """
Usage:
List shapes-file contents as MD5 hash:
list <shapes input>
Create patch:
extract <shapes input> <shapes patch (out)> <collection numbers: 0 to 31>
Apply patch:
patch <shapes input> <shapes patch (in)> <shapes output (separate file)>
"""

if len(sys.argv) <= 1:
	print usage
	sys.exit(0)


# Simple exception class that makes a message
class MyException(Exception):
	def __init__(self,msg):
		self.msg = msg
	def __str__(self):
		return repr(self.msg)

# Reads shapes into list of 32 lists of all collections chunks (one for each resolution)
def GetShapes(filename):
	f = open(filename,"rb")
	
	# Get the headers first
	# An array of 32 collection headers
	# A collection header is an array of 2 chunk headers
	# A chunk header is [offset, length]
	header = []
	for c in xrange(32):
		data = struct.unpack(">hH 4l 12s",f.read(32))
		collhdr = [data[2:4],data[4:6]]
		header.append(collhdr)
	
	# Get the shape chunks
	shapes = []
	for collhdr in header:
		collection = []
		for chkhdr in collhdr:
			offset = chkhdr[0]
			length = chkhdr[1]
			if offset >= 0:
				f.seek(offset)
				chunk = f.read(length)
				# Silently fail if cut off at end of file
				if len(chunk) < length:
					chunk = None
			else:
				chunk = None
			collection.append(chunk)
		shapes.append(collection)
	
	f.close()
	return shapes

def PutShapes(filename,shapes):
	f = open(filename,"wb")
	
	# Write an initial blank header,
	# then write the chunks and return to overwrite
	# the header
	BlankHeader = (32*32)*'\0'
	f.write(BlankHeader)
	
	# Write the chunks
	header = []
	for collection in shapes:
		collhdr = []
		for chunk in collection:
			if chunk != None:
				offset = f.tell()
				length = len(chunk)
				f.write(chunk)
			else:
				offset = -1
				length = 0
			chkhdr = [offset,length]
			collhdr.append(chkhdr)
		header.append(collhdr)
	
	# Write the real header
	Padding = 12*'\0'
	f.seek(0)
	for collhdr in header:
		data = struct.pack(">hH 4l 12s", 0, 0,
			collhdr[0][0], collhdr[0][1], collhdr[1][0], collhdr[1][1],
				Padding)
		f.write(data)
		
	f.close()
	
	# Set OSX type and creator codes using its developer-tools SetFile program
	# If it is not found, then skip over it
	# Should skip over it if not OSX
	# Set here to Moo shapes type code and Aleph One creator code
	
	OSX_SetFile = "/Developer/Tools/SetFile"
	if os.access(OSX_SetFile,os.X_OK):
		os.system("%s -t shp\xb0 -c 26.A %s" % (OSX_SetFile, pipes.quote(filename)))

cmd = sys.argv[1].lower()

if cmd == "list":
	shapes = GetShapes(sys.argv[2])
	print '\t'.join(["Collection","Depth","Length","MD5 Hash"])
	for c, collection in enumerate(shapes):
		for d, chunk in enumerate(collection):
			outlist = [c, d]
			if chunk != None:
				outlist += [len(chunk), hashlib.md5(chunk).hexdigest()]
			else:
				outlist += [-1, ""]
			print '\t'.join(map(str,outlist))
	
elif cmd == "extract":
	
	# Get which ones to export (0 to 31)
	to_export = map(int,sys.argv[4:])
	
	export_mask = 32*[False]
	for c in to_export: export_mask[c] = True
	
	# Go!
	shapes = GetShapes(sys.argv[2])
	for c, em in enumerate(export_mask):
		if not em:
			shapes[c] = 2*[None]
	PutShapes(sys.argv[3],shapes)

elif cmd == "patch":

	orig_shapes = GetShapes(sys.argv[2])
	patch_shapes = GetShapes(sys.argv[3])
	
	shapes = []
	for c in xrange(32):
		collection = []
		for d in xrange(2):
			patch_chunk = patch_shapes[c][d]
			if patch_chunk != None:
				chunk = patch_chunk
			else:
				chunk = orig_shapes[c][d]
			collection.append(chunk)
		shapes.append(collection)
	
	PutShapes(sys.argv[4],shapes)
